local astar_handlers = include("sim/astar_handlers")
local util = include( "modules/util" )
local astar = include("modules/astar")
local simdefs = include("sim/simdefs")
local guarddefs = include( "sim/unitdefs/guarddefs" )
local mathutil = include( "modules/mathutil" )
local simquery = include( "sim/simquery" )

function astar_handlers.aihandler:_handleNode(to_cell, from_node, goal_cell)
	local emCost = mathutil.dist2d( to_cell.x, to_cell.y, goal_cell.x, goal_cell.y )
	local dc = simquery.getMoveCost( from_node.location, to_cell )

	-- Since we are soft-pathing, we may have dynamic impass in the way.  Handle this by
	-- adding a fixed cost per obstacle.
	for i,unit in ipairs(to_cell.units) do
		if unit ~= self._unit and unit:getPlayerOwner() == self._unit:getPlayerOwner() and unit:hasTrait("dynamicImpass") then
			--dc = dc + 2
			emCost = emCost + 2
			break
		end
	end

	-- Fetch a Node for the given location and set its parameters
	local n = self:getNode(to_cell, from_node)
	n.mCost = from_node.mCost + dc
	n.score = n.mCost + emCost
    
	return n
end

local oldCoop_handler = astar_handlers.coop_handler

local AStarHandler_Coop = {}

function AStarHandler_Coop:new(unit, heuristicFn, ...)
	if guarddefs["ce_dummy_pather"] then
		local instance = util.tcopy( self )
		instance._unit = unit
		instance._heuristicFn = heuristicFn
		instance._sim = unit:getSim()
		instance._fromNode = unit:getLocation()
		instance._unitPenalties = {}
		instance._boardw = instance._sim:getBoardSize() -- only need width.
		instance._bestEC = math.huge
		
		if unit:getTraits().vip then
			for i,v in ipairs(unit:getBrain():getSenses().interests) do
				if v.reason == simdefs.REASON_LOSTTARGET then
					instance.lostTargetInterest = v
				end
			end
		end
		
		return instance
	else
		return oldCoop_handler:new(unit, heuristicFn, ...)
	end
end

function AStarHandler_Coop:getNodeLid( x, y, t )
	return self._unit:getPather():getNodeLid(x, y, t)
end

function AStarHandler_Coop:getNode( cell, t, parentNode )
	if cell then
		self.nc = (self.nc or 0) + 1
		t = t or 0
		local lid = self:getNodeLid( cell.x, cell.y, t )
		local node = astar.Node:new( cell, 0, lid, parentNode )
		node.t = t
		node.timesWaited = parentNode and parentNode.timesWaited or 0
		node.rawCost = 0
		return node
	end
end


function AStarHandler_Coop:getAdjacentNodes( cur_node, goal_cell, closed )
	-- Given a node, return a table containing all adjacent nodes
	local result = {}	
	local cell = cur_node.location
    local sim = self._sim
	local simquery = sim:getQuery()
	local n = nil
	local reason = nil

	-- Can also 'pause': eg. stay at same place with advanced t.
	n = self:_handleNode( cell, cur_node, goal_cell )
	if n then
		table.insert( result, n )
	end

	for dx = -1,1 do
		for dy = -1,1 do
			local target_cell = sim:getCell( cell.x + dx, cell.y + dy )
			local target_lid = self:getNodeLid( cell.x + dx, cell.y + dy, cur_node.t + 1 )
			if target_cell and not (dx == 0 and dy == 0) and not closed[ target_lid ] then
				local canPath = simquery.canSoftPath( sim, self._unit, cell, target_cell )
				if self.lostTargetInterest and self.lostTargetInterest.x == (cell.x + dx) and self.lostTargetInterest.y == (cell.y + dy) then
					canPath = false
				end
				if canPath then
                    n, reason = self:_handleNode( target_cell, cur_node, goal_cell )
					
					if n then
						table.insert(result, n)
					end
				end
			end
		end
	end

	return result
end

-- There's some fine tuning that can be done with this value
local MAX_WAIT = 8

function AStarHandler_Coop:isGoal( node, goal )
	local maxWaitTime = MAX_WAIT
	local sim = self._sim
	local goalCell = sim:getCell( goal.x, goal.y )
	local dx, dy = goal.x - node.location.x, goal.y - node.location.y
	
	if dx == 0 and dy == 0 then
		maxWaitTime = 0
	elseif goalCell and goalCell.impass > 0 then
		local exit
		if dx == 1 and dy == 0 then
			exit = node.location.exits[ simdefs.DIR_E ]
		elseif dx == -1 and dy == 0 then
			exit = node.location.exits[ simdefs.DIR_W ]
		elseif dx == 0 and dy == 1 then
			exit = node.location.exits[ simdefs.DIR_N ]
		elseif dx == 0 and dy == -1 then
			exit = node.location.exits[ simdefs.DIR_S ]
		end
		
		if exit and not (exit.locked and exit.closed) then
			maxWaitTime = 3
		end
	else
		maxWaitTime = maxWaitTime - math.max(3 + self._bestEC - node.ec, 0)
	end
	
	if node.timesWaited + node.rawCost >= self:getMP() + maxWaitTime then
		--log:write(string.format("%d maybe found goal at (%d, %d, %d)", self._unit:getID(), node.location.x, node.location.y, node.t ))
		
		local pather = self._unit:getPather()
		if pather:checkPathReservation( self:getNodeLid( node.location.x, node.location.y, -1 ) ) then
			--log:write(string.format("%d can't stop at (%d, %d, %d) because %d is already stopping there", self._unit:getID(), node.location.x, node.location.y, node.t, pather:checkPathReservation( self:getNodeLid( node.location.x, node.location.y, -1 ) ).path.unit:getID() ))
			return false
		end
		
		local cellLid = self:getNodeLid( node.location.x, node.location.y, 0 )
		local reserves = pather._cellReserves[cellLid]
		if reserves and #reserves > 0 and reserves[#reserves] > node.t then
			for i = node.t + 1, reserves[#reserves] do
				if pather:checkPathReservation( self:getNodeLid( node.location.x, node.location.y, i ) ) then
					--log:write(string.format("%d can't stop at (%d, %d, %d) because %d will be there at t = %d", self._unit:getID(), node.location.x, node.location.y, node.t, pather:checkPathReservation( self:getNodeLid( node.location.x, node.location.y, i ) ).path.unit:getID(), i ))
					return false
				end
			end
		end
		
		return true
	end
	
	return false
end

function AStarHandler_Coop:getMP()
	if self._sim:getCurrentPlayer():isPC() or self._sim:getTurnState() ~= simdefs.TURN_PLAYING or self._unit:getTraits().mp < 1.5 then
		return self._unit:getTraits().mpMax
	else
		return self._unit:getTraits().mp
	end
end

function AStarHandler_Coop:_handleNode(to_cell, from_node, goal_cell)
	if self._unit:getTraits().dynamicImpass then
		if self._unit:getPather():checkPathReservation( self:getNodeLid( to_cell.x, to_cell.y, from_node.t + 1 ) ) then
			--log:write( "RESERVED at (%.d, %.d, t = %.d) by %.d", to_cell.x, to_cell.y, from_node.t + 1, self._unit:getPather():checkPathReservation( self:getNodeLid( to_cell.x, to_cell.y, from_node.t + 1 ) ).path.unit:getID() )
			return nil
		elseif self._unit:getPather():checkPathReservation( from_node.lid + 1 ) and self._unit:getPather():checkPathReservation( self:getNodeLid(to_cell.x, to_cell.y, from_node.t) ) then
			--log:write( "HEAD-ON MOVE, ILLEGAL at (%.d, %.d, t = %.d)", from_node.location.x, from_node.location.y, from_node.t + 1 )
			return nil
		elseif self._unit:getPather():checkPathReservation( self:getNodeLid( to_cell.x, to_cell.y, -1 ) ) and self._unit:getPather():checkPathReservation( self:getNodeLid( to_cell.x, to_cell.y, -1 ) ).node.t <= from_node.t + 1 then
			--local reservedNode = self._unit:getPather():checkPathReservation( self:getNodeLid( to_cell.x, to_cell.y, -1 ) )
			--log:write( "%d endpoint reserved at (%.d, %.d, t = %.d) by %d (%.d, %.d, t = %.d)", self._unit:getID(), to_cell.x, to_cell.y, from_node.t + 1, reservedNode.path.unit:getID(), reservedNode.node.location.x, reservedNode.node.location.y, reservedNode.node.t )
			return nil
		end
	end
	
	local timesWaited = from_node.timesWaited
	
	-- Direct cost.
	local dc = simquery.getMoveCost( from_node.location, to_cell )
	local wc = 0
	if from_node.rawCost + dc > self:getMP() then
		return nil
	end
	
	if to_cell.x == from_node.location.x and to_cell.y == from_node.location.y then
		timesWaited = timesWaited + 1
		local maxWaitTime = self:getMP() + MAX_WAIT
		
		if to_cell.x == goal_cell.x and to_cell.y == goal_cell.y then
			-- Waiting on the goal cell must cost 0
		elseif from_node.rawCost + dc > self:getMP() - 1 then
			-- The same is true for waiting after using all our AP, which in a sense is also 'the goal'
		else
			-- This value is hard to tune: If it's too high guards will take inefficient paths instead of taking turns, if it's too low guards can stop pathing too early because they think waiting is the best choice
			wc = 0.2
		end
		
		if timesWaited > maxWaitTime then
			return nil
		end
	else
		-- Resetting the wait counter when we move allows us to be flexible about dividing up our moves, while still enforcing a strictly finite search depth
		timesWaited = 0
	end
	
	-- Estimated cost
	local ec = self._heuristicFn( to_cell, from_node, goal_cell )
	if ec == nil then
		return nil -- Missing heuristic implies that this node is not pathable.
	end
	
	if ec < self._bestEC then
		self._bestEC = ec
	end
	
	local penalty = 0
	if self._unit:getTraits().dynamicImpass then
		for i,unit in ipairs(to_cell.units) do
			if self._unitPenalties[unit] then
				penalty = penalty + self._unitPenalties[unit]
			else
				local addPenalty = 0
				if unit ~= self._unit and unit:hasTrait("dynamicImpass") then
					if unit:getPlayerOwner() == self._unit:getPlayerOwner() then
						if (unit:getBrain() and unit:getBrain():getTarget()) then
							addPenalty = 6
						elseif self._sim:getCurrentPlayer():isNPC() and unit:getMP() < 1.5 then
							addPenalty = 2
						elseif self._unit:getPather():checkPathReservation( self:getNodeLid( to_cell.x, to_cell.y, -2 ) ) then
							addPenalty = 1.5
						else
							addPenalty = 0.1
						end
					elseif simquery.isUnitUnderOverwatch( unit ) and from_node.location.x ~= to_cell.x and from_node.location.y ~= to_cell.y then
						addPenalty = 5
					end
				end
				
				penalty = penalty + addPenalty
				self._unitPenalties[unit] = addPenalty
			end
		end
	end
	
	--penalty = penalty + wc

	-- Fetch a Node for the given location and set its parameters
	local n = self:getNode(to_cell, from_node.t + 1, from_node)
	n.rawCost = from_node.rawCost + dc
	n.ec = ec
	--n.penalty = from_node.penalty + penalty
	n.mCost = from_node.mCost + dc + wc + 0.25 * penalty
	n.score = n.mCost + penalty + ec * 1.001 -- Break ties by encouraging forward movement
	n.timesWaited = timesWaited

	return n
end

astar_handlers.coop_handler = AStarHandler_Coop