local IdleSituation = include("sim/btree/situations/idle")
local simquery = include( "sim/simquery" )
local simdefs = include( "sim/simdefs" )
local cdefs = include( "client_defs" )
local unitdefs = include( "sim/unitdefs" )
local util = include( "modules/util" )
local mathutil = include( "modules/mathutil" )


local function canGuardSpawn( sim, unit, x, y )
	local cell = sim:getCell( x, y )
	if cell.tileIndex == nil or cell.tileIndex == cdefs.TILE_SOLID or cell.exitID or cell.cell ~= nil or cell.impass > 0 then
		return false
	end

	if simquery.cellHasTag(sim, cell, "noguard") then        
		return false
	end

    if not simquery.canPath( sim, unit, nil, cell ) then
		return false
	end

	return true
end

local function findGuardSpawn( sim, room, unit )
    local cells = {}
    for _, rect in ipairs(room.rects) do
        for x = rect.x0, rect.x1 do
            for y = rect.y0, rect.y1 do
                if canGuardSpawn( sim, unit, x, y ) then
                    table.insert( cells, x )
                    table.insert( cells, y )
                end
            end
        end
    end

    if #cells > 0 then
        local i = sim:nextRand( 1, #cells / 2 )
        return cells[2 * i - 1], cells[2 * i]
    end
end

local function markPatrolRoom( sim, room, unit )
    local roomData = sim._mutableRooms[ room.roomIndex ]
    if roomData.patrolUnits == nil then
        roomData.patrolUnits = {}
    end
    table.insert( roomData.patrolUnits, unit:getID() )
end

local function canPatrolRoom( sim, startRoom, room, unit )
    local beginnerPatrols = sim:getParams().difficultyOptions.beginnerPatrols
    if beginnerPatrols then
        -- Beginning patrols MUST start and end in the same 'zone'.
        return startRoom.zoneID == room.zoneID
    else
        -- Removed this restriction
		
        --if startRoom == room  then
            --return false
        --end

        if room.tags.entry then
            local roomData = sim._mutableRooms[ room.roomIndex ]
            local currentCount = roomData.v and #roomData.patrolUnits or 0
            if currentCount >= 1 then
                return false
            end
        end
    end
    return true
end

function IdleSituation:checkPatrolNumbers(sim)
	local p = sim._patrols
	if not p then
		p = {
			guards = 0,
			patrols = 0,
			tier2 = 0,
		}
		
		for i, unit in pairs(sim:getNPC():getUnits()) do
			if unit:getBrain() and not unit:getTraits().reconPatrol and (unit:getBrain():getSituation() == nil or unit:getBrain():getSituation().ClassType == simdefs.SITUATION_IDLE) then
				if unit:getTraits().patrolPath or unit:getTraits().nopatrol then
					--Predetermined patrol, we count this as a regular patrol and therefore we subtract one
					p.patrols = p.patrols - 0.5
					p.tier2 = p.tier2 + 0.5
				elseif not unit:getTraits().reconPatrol then
					--Count the number of patrols of each type to make
					p.guards = p.guards + 1
					p.patrols = p.patrols + 0.5
					p.tier2 = p.tier2 + 0.5
				end
			end
		end
		
		sim._patrols = p
	else
		for i, unit in pairs(sim:getNPC():getUnits()) do
			if unit:getBrain() and not unit:getTraits().reconPatrol and (unit:getBrain():getSituation() == nil or unit:getBrain():getSituation().ClassType == simdefs.SITUATION_IDLE) then
				--Count the number of patrols of each type to make
				p.guards = p.guards + 1
				p.patrols = p.patrols + 0.5
				p.tier2 = p.tier2 + 0.5
			end
		end
	end
	
	return p
end

local function flood( sim, unit, start_cell, range, minRange )
	sim:dispatchEvent("NO EVENT")

	local open_list = { {start_cell, 0} }
	local closed_indexed = { [start_cell] = 0 }
	local closed_list = { start_cell }
	
	local i = 1
	while i <= #open_list do
		local cell, cost = open_list[i][1], open_list[i][2]
	
		if cost == closed_indexed[cell] then--Check in case we've been added multiple times
			for dx = -1,1 do
				for dy = -1,1 do
					local target_cell = sim:getCell( cell.x + dx, cell.y + dy )
					if target_cell and cell ~= target_cell and simquery.canSoftPath( sim, unit, cell, target_cell ) then
						local target_cost = cost + simquery.getMoveCost( cell, target_cell )
						
						if target_cost <= range then
							if not closed_indexed[target_cell] or closed_indexed[target_cell] > target_cost then
								if not closed_indexed[target_cell] then
									table.insert(closed_list,target_cell)
								end
								closed_indexed[target_cell] = target_cost
								table.insert(open_list,{target_cell, target_cost})
							end
						end
					end
				end
			end
		end
		
		i = i + 1
	end
	
	local output = {}
	local outputIdx = {}
	
	for i, cell in ipairs(closed_list) do
		local cost = closed_indexed[cell]
		if cost >= minRange and canPatrolRoom( sim, start_cell.procgenRoom, cell.procgenRoom, unit ) then
			table.insert(output,{cell,cost})
			outputIdx[cell.id] = #output
		end
	end
	
	return output, outputIdx
end

local oldGeneratePatrolPath = IdleSituation.generatePatrolPath

function IdleSituation:generatePatrolPath( unit, x0, y0, noPatrolCheck )
    assert( unit:getBrain():getSituation() == self )
	
	local sim = unit:getSim()
	
	if unitdefs.lookupTemplate("ce_dummy_patrol") then
		if not sim.ce_shouldGeneratePatrols then
			return
		end
		
		--This sets up so that we get an even amount of regular and advanced patrols
		local p = sim._patrols
		if p == nil or p.guards == 0 then
			p = self:checkPatrolNumbers(sim)
		end
		
		-- make a patrol!
		if sim._rooms == nil then -- no patrols for pre-designed levels.
			return
		end
		
		if (x0 == nil or y0 == nil) then
			local i = sim:nextRand( 1, #sim._rooms )
			local room = sim._rooms[ i ]
			x0, y0 = findGuardSpawn( sim, room, unit )
			if x0 == nil or y0 == nil then
				local rooms = util.tdupe( sim._rooms )
				while #rooms > 1 do
					table.remove(rooms, i)
					i = sim:nextRand( 1, #rooms )
					room = rooms[ i ]
					x0, y0 = findGuardSpawn( sim, room, unit )
					if x0 and y0 then
						break
					end
				end
			end
		end
		
		local cell = sim:getCell( x0, y0 )
		if cell then
			local startRoom = cell.procgenRoom
			local path = {}
			table.insert( path, { x = x0, y = y0 } )
			markPatrolRoom( sim, startRoom, unit )
			
			local maxMP, triplepoint = unit:getTraits().mpMax, false
			local minDist = maxMP - 1--The minimum distance between patrol points
			local targetScore--The optimal length of the whole patrol in the case of a triple patrol (search ends early once a patrol of this length is found)
			if sim:getParams().difficultyOptions.beginnerPatrols then
				maxMP, minDist = 4, 0
			end
			
			--Determines how each type of patrol is rounded. 0.5 means round down, 0 means round up.
			local ROUND_TIER2 = 0
			local ROUND_STANDARD = 0.5
			
			--Determine what kind of patrol we are making
			if p.tier2 > ROUND_TIER2 or unit:getTraits().reconPatrol then
				local odds = 4--odds of getting a long patrol
				if p.patrols <= ROUND_STANDARD then
					--Only tier 2 patrols remain
					odds = 2
				end
			
				if not unit:getTraits().reconPatrol and sim:nextRand(odds) == 1 then
					--Long Patrol
					maxMP = math.floor(maxMP * 1.8)
					minDist = maxMP - 0.9
					p.tier2 = p.tier2 - 1
					p.guards = p.guards - 1
				elseif unit:getTraits().reconPatrol or p.patrols <= ROUND_STANDARD or sim:nextRand(2) == 1 then
					--Triple Patrol
					triplepoint = true
					if unit:getTraits().reconPatrol then
						maxMP = math.max(maxMP,12+(unit:getTraits().hardmode and 3 or 0))
						--Recon Patrols don't count so don't subtract from p.guards!
					else
						p.guards = p.guards - 1
						p.tier2 = p.tier2 - 1
					end
					
					targetScore = maxMP * 3 - 2
					minDist = maxMP - (unit:getTraits().threePointRange or 7)
				else
					--Regular Patrol
					p.guards = p.guards - 1
					p.patrols = p.patrols - 1
				end
			else
				--Regular Patrol
				p.guards = p.guards - 1
				p.patrols = p.patrols - 1
			end
			
			local pool, poolIdx = flood( sim, unit, cell, maxMP, minDist )
			
			if #pool == 0 then
				pool, poolIdx = flood( sim, unit, cell, minDist, 0 )
			end
			
			unit:getTraits().patrolPool = pool

			
			if triplepoint then
				--Triple Patrol
				local pool1 = util.tdupe(pool)
				
				local bestCell1, bestCell2
				local bestScore = 0
				
				while #pool1 > 0 do
					local cell1 = table.remove( pool1, sim:nextRand(#pool1) )
					
					local pool2 = flood( sim, unit, cell1[1], maxMP, minDist )
					
					for i, cell2 in ipairs(pool2) do
						if poolIdx[cell2[1].id] then
							local score = cell2[2] + pool[poolIdx[cell2[1].id]][2] + pool[poolIdx[cell1[1].id]][2]
							
							if score > bestScore then
								bestCell1 = cell1
								bestCell2 = cell2
								bestScore = score
							end
						end
					end
					
					if bestScore >= targetScore then
						break
					end
				end
				
				if bestCell1 then
					path[1].l = bestCell1[2]
				
					table.insert( path, { x = bestCell1[1].x, y = bestCell1[1].y, l = bestCell2[2] } )
					markPatrolRoom( sim, bestCell1[1].procgenRoom, unit )
					
					table.insert( path, { x = bestCell2[1].x, y = bestCell2[1].y, l = pool[poolIdx[bestCell2[1].id]][2] } )
					markPatrolRoom( sim, bestCell2[1].procgenRoom, unit )
				else
					p.tier2 = p.tier2 + 1
					p.patrols = p.patrols - 1
				end
			end
			
			--Regular Patrol
			if #path <= 1 then
				local bestCell
				local bestScore = 0
				for i,cell in ipairs( pool ) do
					local score = mathutil.distSqr2d(x0,y0,cell[1].x,cell[1].y)
					if score > bestScore then
						bestCell = cell
						bestScore = score
					end
				end
				
				if bestCell then
					path[1].l = bestCell[2]
					table.insert( path, { x = bestCell[1].x, y = bestCell[1].y, l = bestCell[2] } )
					markPatrolRoom( sim, bestCell[1].procgenRoom, unit )
				end
			end
			--Loop patrols for triple
			unit:getTraits().patrolmode = simdefs.PATROL_LOOPING
			unit:getTraits().nextWaypoint = 2
			unit:getTraits().patrolPath = path
		else
			unit:getTraits().patrolPath = nil
		end
	else
		oldGeneratePatrolPath( self, unit, x0, y0, noPatrolCheck )
	end
end

local oldgenerateStationaryPath = IdleSituation.generateStationaryPath

function IdleSituation:generateStationaryPath( unit, x0, y0)
	local sim = unit:getSim()
	
	if unitdefs.lookupTemplate("ce_dummy_patrol") then
		if not sim.ce_shouldGeneratePatrols then
			return
		end
	
		--This sets up so that we get an even amount of regular and advanced patrols
		local p = sim._patrols
		if p == nil or p.guards == 0 then
			p = self:checkPatrolNumbers(sim)
		end

		if unit:getTraits().always_patrol or unit:getTraits().reconPatrol or p.patrols <= 0 then
			self:generatePatrolPath( unit, x0, y0 )
			return
		end
		
		p.guards = p.guards - 1
		p.patrols = p.patrols - 1
	end
		
	oldgenerateStationaryPath( self, unit, x0, y0)
end
