local array = include( "modules/array" )
local util = include( "modules/util" )
local simdefs = include( "sim/simdefs" )
local simquery = include( "sim/simquery" )
local line_of_sight = include( "sim/line_of_sight" )
local unitdefs = include("sim/unitdefs")

local RAYCAST_PERCENT = 0.45

local function proj(x1, y1, x2, y2)
	--Rotate 90 degrees then take the dot product
	return -x1*y2 + y1*x2
end

local function clockwise(x1, y1, x2, y2)
	return -x1*y2 + y1*x2 > 0
end

local function isPointInArc(x, y, x0, y0, arc0X, arc0Y, arc1X, arc1Y, distance)
	local vx = x - x0
	local vy = y - y0
	return clockwise(arc1X,arc1Y,vx,vy) and not clockwise(arc0X,arc0Y,vx,vy) and (vx * vx + vy * vy) <= distance * distance
end

local function floodLOS( sim, start_cell, facing, halfArc, distance, dx, dy, ignoresWalls )
	local x0, y0 = start_cell.x + (dx or 0), start_cell.y + (dy or 0)
	local open_list = { [start_cell.id ] = start_cell }
	local close_list = { [start_cell.id ] = start_cell }
    local vis_list = { [start_cell.id ] = start_cell }
	
	local quarterArc = halfArc / 2
	--local loop = 1
	while not util.tempty( open_list ) do
		local cellID, cell = next( open_list )
		open_list[ cell.id ] = nil
		
		for i = 1, #simdefs.OFFSET_NEIGHBOURS, 2 do
			local dx, dy = simdefs.OFFSET_NEIGHBOURS[i], simdefs.OFFSET_NEIGHBOURS[i+1]
			local adj_cell = sim:getCell( cell.x + dx, cell.y + dy )
			local ok = adj_cell ~= nil
			ok = ok and open_list[ adj_cell.id ] == nil
			ok = ok and close_list[ adj_cell.id ] == nil
			--ok = ok and math.abs(mathutil.angleDiff( facing, mathutil.atan2t( adj_cell.x - x0, adj_cell.y - y0 ))) <= halfArc
			if ok then
                if ignoresWalls then
					--Manual los calculation!!!
					
					local arc0X = math.cos(facing - halfArc)
					local arc0Y = math.sin(facing - halfArc)
					
					local arc1X = math.cos(facing + halfArc)
					local arc1Y = math.sin(facing + halfArc)
					
					if isPointInArc(adj_cell.x, adj_cell.y, x0, y0, arc0X, arc0Y, arc1X, arc1Y, distance) then
						vis_list[ adj_cell.id ] = adj_cell
					end
					
					open_list[ adj_cell.id ] = adj_cell
					close_list[ adj_cell.id ] = adj_cell
				else
				
					local percentSamples = sim:getLOS().losmap:raycastToCell( x0, y0, adj_cell.x, adj_cell.y, facing - quarterArc, quarterArc , distance )
					percentSamples = percentSamples + sim:getLOS().losmap:raycastToCell( x0, y0, adj_cell.x, adj_cell.y, facing + quarterArc, quarterArc , distance )
					if percentSamples > 0 then
						open_list[ adj_cell.id ] = adj_cell
						if percentSamples >= RAYCAST_PERCENT then
							vis_list[ adj_cell.id ] = adj_cell
						end
					end
					close_list[ adj_cell.id ] = adj_cell
					
				end
			end
		end
	end

	return vis_list
end

function line_of_sight:calculateUnitLOS( start_cell, unit )
	if unit:getTraits().globalVision then
		self.sim:getTags().globalVision = self.sim:getTags().globalVision or 0
		self.sim:getTags().globalVisionUnits = self.sim:getTags().globalVisionUnits or {}
		if unit:getTraits().hasSight then
			local cells = {}
			self.sim:forEachCell( function(cell)
				cells[cell.id] = cell
			end)
			for i, unit in pairs(self.sim:getAllUnits()) do
				unit:getTraits().tagged = true
				unit:getTraits().patrolObserved = true
			end
			if not self.sim:getTags().globalVisionUnits[unit] then
				self.sim:getTags().globalVisionUnits[unit] = true
				self.sim:getTags().globalVision = self.sim:getTags().globalVision + 1
			end
			self.sim:dispatchEvent( simdefs.EV_EXIT_MODIFIED )
			return cells
		else
			if self.sim:getTags().globalVisionUnits[unit] then
				self.sim:getTags().globalVisionUnits[unit] = nil
				self.sim:getTags().globalVision = self.sim:getTags().globalVision - 1
			end
			self.sim:dispatchEvent( simdefs.EV_EXIT_MODIFIED )
			return {}
		end
	end
    local halfArc = simquery.getLOSArc( unit ) / 2
	if halfArc > math.pi then
		unit:getTraits().LOSarc = math.pi * 2
		halfArc = math.pi
	end
    local distance = unit:getTraits().LOSrange
	local facingRads = unit:getFacingRad()

    local dx, dy
    if self.sim:isVersion( "0.17.2" ) then
        -- Alter camera raycasting to match the viz (corner cameras actually source from the corner of their cell, argh).
        if unit:getTraits().mainframe_camera then
            local facing = unit:getFacing()
            if facing % 2 == 1 then
                dx, dy = simquery.getDeltaFromDirection( facing )
                if self.sim:isVersion( "0.17.3" ) then
                    dx, dy = -dx/2 * 0.9999, -dy/2 * 0.9999
                else
                    dx, dy = -dx/2 + 0.00001, -dy/2 + 0.00001
                end
            end
        end
    end

	local cells = floodLOS( self.sim, start_cell, facingRads, halfArc, distance, dx, dy, unit:getTraits().ignoresWalls )

    local facing = unit:getFacing()
	
	if unit:getTraits().LOSrads == nil and facing % 2 == 1 and (unit:getTraits().LOSrange == nil or unit:getTraits().LOSrange >= 1 or not unitdefs.lookupTemplate("ce_dummy_pather")) then
        -- MAGICAL SIGHT.  On a diagonal facing, see the adjacent two cells.
        local exit1 = start_cell.exits[ (facing + 1) % simdefs.DIR_MAX ] 
        if simquery.isOpenExit( exit1 ) then
            cells[ simquery.toCellID( exit1.cell.x, exit1.cell.y ) ] = exit1.cell
        end
        local exit2 = start_cell.exits[ (facing - 1) % simdefs.DIR_MAX ]
        if simquery.isOpenExit( exit2 ) then
            cells[ simquery.toCellID( exit2.cell.x, exit2.cell.y ) ] = exit2.cell
        end

    elseif unit:getTraits().LOSarc and unit:getTraits().LOSarc >= 2 * math.pi and (unit:getTraits().LOSrange or 1) >= 1 then
        for i, dir in ipairs( simdefs.DIR_SIDES ) do
            local exit1 = start_cell.exits[ dir ]
            if simquery.isOpenExit( exit1 ) then
                cells[ simquery.toCellID( exit1.cell.x, exit1.cell.y ) ] = exit1.cell
            end
        end
    end
	
	if unit:getTraits().ignoresWalls then
		unit:getTraits().vizCells = {}
		
		for i, cell in pairs(cells) do
			table.insert(unit:getTraits().vizCells,cell.x)
			table.insert(unit:getTraits().vizCells,cell.y)
		end
	end
	
    return cells
end

function line_of_sight:getVizCells( unitID, vizcells )
	local unit = self.sim:getUnit(unitID)
	
	if unit and unit:getTraits().ignoresWalls then
		for i = 1, #unit:getTraits().vizCells do
			vizcells[i] = unit:getTraits().vizCells[i]
		end
	else
		self.losmap:getVizCells( unitID, vizcells )
	end
end

function line_of_sight:getSeers( x, y )
	local ret = self.losmap:getSeers( x, y )
	
	for i, unit in pairs(self.sim:getNPC():getUnits()) do
		if unit:getTraits().ignoresWalls and not array.find(ret,unit:getID()) and self:hasSight(unit,x,y) then
			table.insert(ret,unit:getID())
		end
	end
	
	return ret
end

function line_of_sight:hasSight( unit, x, y )
	if unit:isValid() then
		if unit:getTraits().ignoresWalls then
			local halfArc = simquery.getLOSArc( unit ) / 2
			local distance = unit:getTraits().LOSrange
			local x0, y0 = unit:getLocation()
			local facing = unit:getFacingRad()
			local arc0X = math.cos(facing - halfArc)
			local arc0Y = math.sin(facing - halfArc)
			local arc1X = math.cos(facing + halfArc)
			local arc1Y = math.sin(facing + halfArc)
		
			return isPointInArc(x, y, x0, y0, arc0X, arc0Y, arc1X, arc1Y, distance)
		else
			return self.losmap:hasSight( unit:getID(), x, y )
		end
	end
end

function line_of_sight:calculatePeripheralLOS( start_cell, unit )
	if unit:getTraits().LOSperipheralArc > math.pi * 2 then
		unit:getTraits().LOSperipheralArc = math.pi * 2
	end
    local halfArc = unit:getTraits().LOSperipheralArc / 2
    local distance = unit:getTraits().LOSperipheralRange
	local facingRads = unit:getFacingRad()
    return floodLOS( self.sim, start_cell, facingRads, halfArc, distance )
end

function line_of_sight:calculateLOS( start_cell, facingRads, halfArc, distance )
	return floodLOS( self.sim, start_cell, facingRads, halfArc, distance )
end