local Senses = include("sim/btree/senses")
local array = include("modules/array")
local util = include( "modules/util" )
local simdefs = include( "sim/simdefs" )
local simquery = include( "sim/simquery" )
local mathutil = include( "modules/mathutil" )
local speechdefs = include( "sim/speechdefs" )
local guarddefs = include( "sim/unitdefs/guarddefs" )

local oldOnTrigger = Senses.onTrigger
local oldUpdate = Senses.update
local oldPickBestInterest = Senses.pickBestInterest

function Senses:canTellBuddy(sim,user,target,seerID)
	if user == target then
		return false -- Can't yell at ourselves
	end
	if target:isDown() then
		return false -- No point yelling at someone who's down
	end
	if user:getPlayerOwner() ~= target:getPlayerOwner() or not target:getBrain() then
		return false -- Don't yell at agent controlled units
	end
	if user:getTraits().ce_yelled or user:getBrain() == nil or user:getBrain():getSituation() == nil or user:getBrain():getSituation().ClassType == simdefs.SITUATION_FLEE or target:isAlerted() or not user:isAlerted() then
		return false -- Don't yell if we've already yelled this turn, if we can't think, if we're a VIP, if the target is already alerted, if we're not alerted
	end
	if not target:getTraits().isGuard or not user:getTraits().isGuard then
		return false -- Don't yell at any non-guards
	end
	if user:getTraits().noInterestDistraction or target:getTraits().noInterestDistraction then
		return false -- Do not yell at or from Pulse Drones (focused AI)
	end
	if target:getTraits().ce_yelledAt then
		return false -- Do not yell at someone that's already been yelled at this turn
	end
	if target:getTraits().isDrone and not user:getTraits().isDrone then
		return false -- Non-drones can't yell at drones
	end
	if user:getBrain():getTarget() then
		return false -- Don't yell if we're in combat
	end
	if user:getTraits().tellBuddyCell and user:getTraits().tellBuddyCell.unit == target then
		local x, y = target:getLocation()
		local cell = user:getTraits().tellBuddyCell
		if x == cell.x and y == cell.y then
			return false -- Don't re-yell if the target hasn't moved
		end
	end
	local interest = user:getBrain():getSenses():getRememberedInterest()
	if interest and interest.reason ~= simdefs.REASON_HACKEDDEVICE then
		return false -- Don't yell if we have an important interest (i.e. we're an enforcer spawned to investigate the objective)
	end
	if seerID ~= user:getID() and not array.find( user:getSeenUnits(), target ) then
		return false -- Don't yell if we can't see the target
	end
	if simquery.cellHasTag( sim, sim:getCell( user:getLocation() ) , "guard_spawn" ) or simquery.cellHasTag( sim, sim:getCell( target:getLocation() ) , "guard_spawn" ) then
		return false -- Don't yell into our out of guard elevators (it causes silly behaviour when guards are in the middle of spawning)
	end
	local currentAgent = sim:getNPC():getCurrentAgent()
	if currentAgent ~= user and currentAgent ~= target then
		return false -- Don't yell if it's someone else acting
	end
	return true
end

local function findCommunicationCell( sim, user, target )
	local x0, y0 = user:getLocation()
	local x,y = target:getLocation()
	local targetCell
	if mathutil.distSqr2d(x,y,x0,y0) < 3 then
		-- Already close enough
		targetCell = sim:getCell( x0, y0 )
	else
		-- Test cells in a 3x3 square around the target
		for dx = -1,1 do
		for dy = -1,1 do
			local testCell = sim:getCell( x + dx, y + dy )
			if testCell then
				local dir = simquery.getDirectionFromDelta( testCell.x - x, testCell.y - y )
				if simquery.isOpenExit( testCell.exits[ dir ]) and simquery.canPath( sim, nil, nil, testCell ) then
					if not targetCell then
						targetCell = testCell -- New targetCell
					else
						-- Multiple options: Pick the targetCell that is the closest to us
						local cell1Dist = mathutil.distSqr2d( testCell.x, testCell.y, x0, y0 )
						local cell2Dist = mathutil.distSqr2d( targetCell.x, targetCell.y, x0, y0 )
						if cell1Dist < cell2Dist then
							targetCell = testCell
						end
					end
				end
			end
		end
		end
	end
	
	if not targetCell then
		-- Give up: Literally try to go on top of them
		targetCell = sim:getCell( x, y )
	end
	
	return targetCell
end

function Senses:onTrigger(sim, evType, evData)
	oldOnTrigger(self, sim, evType, evData)
	
	-- Guard communication
	if self.unit and not self:checkDisabled() and (evType == simdefs.TRG_UNIT_APPEARED or evType == simdefs.TRG_UNIT_WARP or evType == simdefs.TRG_UNIT_DISAPPEARED) and sim:getParams().difficultyOptions.ce_tell_buddy and sim:getCurrentPlayer() and sim:getCurrentPlayer():isNPC() then
		local user = self.unit
		local target = evData.unit
		if not self:canTellBuddy(sim,user,target,evData.seerID) then
			return
		end
		
		user:getTraits().ce_buddyToTell = target
		user:getTraits().ce_yelled = true
		
		 -- Find a destination cell that is within 1 tile of the target we're yelling at
		local x0, y0 = user:getLocation()
		local x,y = target:getLocation()
		local targetCell = findCommunicationCell( sim, user, target )
		local facing = simquery.getDirectionFromDelta(x - x0, y - y0)
		user:getTraits().tellBuddyCell = {x = x,y = y,unit = target} -- Remember the current location of the person we're yelling at (so that we can turn to face them later)
		
		-- Add interests to the user and target
		local interest = self:addInterest(targetCell.x, targetCell.y, simdefs.SENSE_SIGHT, simdefs.REASON_TELL_BUDDY, target)
		local interest2 = target:getBrain():getSenses():addInterest(x, y, simdefs.SENSE_RADIO, simdefs.REASON_YELLED_AT, user)
		interest2.facing = (facing + 4) % 8
		
		if sim:getNPC():getCurrentAgent() == target then
			-- The target moved into our sight
			
			-- Since we're not the currently acting unit, we need to manually emit speech
			sim:dispatchEvent( simdefs.EV_UNIT_ALERTED, { unitID = user:getID() } )
			sim:emitSpeech( user, speechdefs.COMBAT_NEWTARGET )
			interest.noticed = true -- Don't double-speak
			interest2.delay = true -- Inject delay between the two guards speaking
		--elseif sim:getNPC():getCurrentAgent() == user then
			-- We moved/peeked and found the target
			-- Speech is handled in Actions.ReactToInterest
		end
	end
end

local oldIsSameInterest = Senses.isSameInterest

function Senses:isSameInterest(interest, other)
	local sim = self.unit:getSim()
	return oldIsSameInterest(self, interest, other) or (interest.reason == simdefs.REASON_FOUNDDRONE and self.unit:getTraits().droneMonitor and interest.reason == other.reason and interest.investigated == other.investigated and interest.sense == other.sense and interest.sourceUnit == other.sourceUnit)
end

function Senses:addInterest(x, y, sense, reason, sourceUnit, ignoreDisguies)
    assert( sense and reason ) -- By the gods, there must be sense and reason! (hehehe)
	
	local interest = nil
	if not self.unit:getTraits().noInterestDistraction or ( (sense == simdefs.SENSE_RADIO and reason == simdefs.REASON_HUNTING))	then
		-- Ghosts need to not get distracted so easily when trying to melee an agent 
		if (self.unit:getTraits().noPeripheralDetection and self.unit:getSim():getCurrentPlayer():isNPC() and (sense == simdefs.SENSE_PERIPHERAL or reason == simdefs.REASON_LOSTTARGET)) then
			return
		end
		
		interest = {x=x, y=y, sense=sense, reason=reason, sourceUnit=sourceUnit}
		
		-- Is this a new interest or updating an existing one?
		for i = #self.interests, 1, -1 do
			local existing = self.interests[i]
			if self:isSameInterest(interest, existing) then
				if existing.x ~= interest.x or existing.y ~= interest.y then
					if reason == simdefs.REASON_SMOKE and sense == simdefs.SENSE_SIGHT then
						--only closer smoke clouds should be preferred
						local x0, y0 = self.unit:getLocation()
						if mathutil.distSqr2d(x0, y0, interest.x, interest.y) < mathutil.distSqr2d(x0, y0, existing.x, existing.y) then
							existing.x = interest.x
							existing.y = interest.y
						end
					else
						existing.x = interest.x
						existing.y = interest.y
					end
					interest = table.remove(self.interests, i)	--remove and return the existing interest so it gets noticed as a newer event
				else
					return existing
				end
			end
		end

		if interest.sourceUnit and interest.sourceUnit:isValid() and self:hasTarget(interest.sourceUnit) then
			--we already know about this agent, don't notice the interest
			interest.noticed = true
		end

		if interest.sourceUnit == self.unit and interest.sense == simdefs.SENSE_RADIO then
			--it's a hunt interest, don't react
			interest.noticed = true
		end

		if interest.reason == simdefs.REASON_FOUNDCORPSE or
		 interest.reason == simdefs.REASON_FOUNDOBJECT or
		 interest.reason == simdefs.REASON_HUNTING or
		 interest.reason == simdefs.REASON_WITNESS or
		 interest.reason == simdefs.REASON_SMOKE or
		 interest.reason == simdefs.REASON_SHARED or
		 interest.reason == simdefs.REASON_KO or
		 interest.reason == simdefs.REASON_REINFORCEMENTS then
			--this is an interest that alerts guards
			interest.alerts = true
		end


		if interest.sense == simdefs.SENSE_SIGHT
		 or interest.reason == simdefs.REASON_TELL_BUDDY
		 or interest.reason == simdefs.REASON_FOUNDCORPSE
		 or interest.reason == simdefs.REASON_FOUNDDRONE
		 or interest.reason == simdefs.REASON_PATROLCHANGED
		 or interest.reason == simdefs.REASON_SCANNED then
			--don't throw grenades at these interests
			interest.grenadeHit = true
		end

		if interest.sense == simdefs.SENSE_HEARING
		 or (interest.sourceUnit and simquery.isEnemyAgent(self.unit:getPlayerOwner(), interest.sourceUnit,ignoreDisguies) )
		 or interest.reason == simdefs.REASON_CAMERA
		 or (interest.alerts and interest.reason ~= simdefs.REASON_HUNTING) then
			interest.alwaysDraw = true
		end

		table.insert(self.interests, interest)
	end
	return interest
end

function Senses:pickBestInterest(...)
	if guarddefs["ce_dummy_pather"] then
		for i,interest in ipairs(self.interests) do
			interest.scanned = nil
		end
	end
	
	return oldPickBestInterest(self,...)
end